@@ -3,9 +3,9 @@ import enum
 import functools
 import importlib
 import os
+
 from dataclasses import dataclass, field
 from typing import Any, Callable, Dict, List, Optional, Union, Tuple, Sequence
-import warnings
 
 import selenium.common.exceptions as selenium_exceptions
 from bs4 import BeautifulSoup
@@ -15,7 +15,7 @@ from selenium.webdriver.remote import webdriver, webelement
 from selenium.webdriver.support import expected_conditions  # As for API
 from selenium.webdriver.support.ui import WebDriverWait  # type: ignore[attr-defined]
 
-from . import internal
+from . import internal, _hypercache
 
 __all__ = [
     "get_browser",
@@ -31,11 +31,12 @@ __all__ = [
 
 
 def Soup(html: str, **kwargs) -> BeautifulSoup:
+    """Convert a string to a BeautifulSoup object"""
     return BeautifulSoup(html, features="html.parser", **kwargs)
 
 
 class BrowserType(enum.Enum):
-    """An enumeration of popular, supported browsers for browser automation"""
+    """An enumeration of supported browsers for :func:`get_browser`"""
 
     CHROME = {"name": "Chrome", "make_headless": lambda x: x.add_argument("--headless")}
     SAFARI = {"name": "Safari", "make_headless": None}
@@ -60,15 +61,24 @@ def get_browser(
     headless: bool = True,
     arguments: Optional[Sequence[str]] = None,
 ) -> webdriver.WebDriver:
-    """Lil' helper function to attempt to get a valid WebDriver"""
+    """Lil' helper function that attempts to get a valid WebDriver"""
     _cache_key = (
-        priority,
+        tuple((enum.name for enum in priority)),
         headless,
         tuple(arguments) if arguments is not None else None,
     )
     if _cache_key in __get_browser_cache:
-        if __get_browser_cache[_cache_key] is None:
+        output = __get_browser_cache[_cache_key]
+        if output is None:
             raise RuntimeError("Could not find any browser that supports your needs")
+        return output
+    try:
+        output = _hypercache.get("web.get_browser", _cache_key)
+        if output is None:
+            raise RuntimeError("Could not find any browser that supports your needs")
+        return output
+    except KeyError:
+        pass
 
     def _(browser_name: BrowserType) -> Optional[webdriver.WebDriver]:
         try:
@@ -91,7 +101,8 @@ def get_browser(
                 options=options,
                 service_log_path=os.devnull,
             )
-            # There is no better way to check than to open and close it...
+            # There is no better way to check validity
+            # than to open and close it...
             browser().close()
         except selenium_exceptions.WebDriverException:
             return None
@@ -101,8 +112,10 @@ def get_browser(
     for browser in priority:
         found = _(browser)
         if found is not None:
+            _hypercache.set("web.get_browser", _cache_key, found)
             __get_browser_cache[_cache_key] = found
             return found
+    _hypercache.set("web.get_browser", _cache_key, None)
     __get_browser_cache[_cache_key] = None
     # TODO: Download geckodriver and use it
     raise RuntimeError("Could not find any browser that supports your needs")
@@ -116,27 +129,40 @@ class Element:
     Should not be instantiated directly but instead with methods like :meth:`BrowserContext.query_selector`
     """
 
-    item: webelement.WebElement
+    webelement: webelement.WebElement
 
     def soup(self) -> BeautifulSoup:
+        """Convert the element into a BeautifulSoup object"""
         return Soup(self["innerHTML"])
 
+    @property
+    @functools.lru_cache()
+    def name(self) -> str:
+        """Name of the element. (e.g. 'a' or 'div')"""
+        return self.webelement.tag_name
+
     # get_attribute documentation says it may return booleans...
     # @functools.lru_cache()
     def get(self, attr: str) -> Optional[str]:
-        return self.item.get_attribute(attr)  # type: ignore
+        """Get the IDL attribute of this element (e.g. `innerHTML`, `value`)"""
+        return self.webelement.get_attribute(attr)  # type: ignore
 
+    # TODO: Use soup and cache soup
+    # To avoid selenium.common.exceptions.StaleElementReferenceException
     def __getitem__(self, attr: str) -> str:
+        """Like :meth:`get` but doesn't return None"""
         output = self.get(attr)
         if output is None:
             raise KeyError(attr)
         return output
 
     def click(self) -> None:
-        self.item.click()  # type: ignore
+        """Simulate a user click on this element"""
+        self.webelement.click()  # type: ignore
 
-    def send_keys(self, keys) -> None:
-        self.item.send_keys(keys)  # type: ignore
+    def type(self, keys: str) -> None:
+        """Simulate a user typing stuff on this element. Examples include filling out a form or text input"""
+        self.webelement.send_keys(keys)  # type: ignore
 
 
 @internal.add_typo_safety
@@ -147,8 +173,17 @@ class BrowserContext:
     _waits: Dict[float, WebDriverWait] = field(default_factory=dict)
 
     def go(self, to: str) -> None:
+        """Change the current URL to `to`"""
         self.driver.get(to)
 
+    def set_url(self, url: str) -> None:
+        """Alias of :meth:`go`"""
+        self.go(url)
+
+    def back(self) -> None:
+        """Go back one page in browser history"""
+        self.driver.back()
+
     def query_selector(self, css_selector: str) -> Element:
         """Instantly find an element that matches the given CSS selector
 
@@ -170,6 +205,21 @@ class BrowserContext:
         return Element(self.driver.find_element(By.CSS_SELECTOR, css_selector))
 
     def css(self, css_selector: str, wait_up_to: float = 10.0) -> Element:
+        """Find something with CSS with a timeout
+
+        Args:
+            css_selector (str): The CSS selector to find an element
+            wait_up_to (float): self-explanatory
+
+        Returns:
+            Element: The element found
+
+        Raises:
+            TimeoutError: Timed out
+
+        See also:
+             :meth:`wait`
+        """
         return self.wait(
             expected_conditions.presence_of_element_located(  # type: ignore
                 (By.CSS_SELECTOR, css_selector)
@@ -180,12 +230,18 @@ class BrowserContext:
     # TODO: css_all
 
     def query_selector_all(self, css_selector: str) -> List[Element]:
+        """:meth:`query_selector` for multiple elements"""
         return list(
             map(Element, self.driver.find_elements(By.CSS_SELECTOR, css_selector))
         )
 
-    def js(self, javascript: str) -> Any:
-        return self.driver.execute_script(javascript)
+    def js(self, javascript: str, *args) -> Any:
+        """Run JavaScript"""
+        return self.driver.execute_script(javascript, *args)
+
+    def run_script(self, javascript: str, *args) -> Any:
+        """Alias for :meth:`js`"""
+        return self.js(javascript, *args)
 
     def _get_wait_up_to(self, seconds: float) -> WebDriverWait:
         if seconds not in self._waits:
@@ -245,16 +301,16 @@ class Browser(contextlib.AbstractContextManager):
     url: str
     # pylint: disable=unnecessary-lambda
     # No, we need that lambda for lazy loading
-    driver: webdriver.WebDriver = field(default_factory=get_browser)
-
-    def __post_init__(self):
-        self.driver = self.driver()
+    driver: Callable[[], webdriver.WebDriver] = field(default_factory=get_browser)
 
     def open(self) -> BrowserContext:
+        """Start the browser"""
+        self.driver = self.driver()  # XXX: Don't make Browsers one-time
         self.driver.get(self.url)
         return BrowserContext(self.url, self.driver)
 
     def close(self) -> None:
+        """Close the browser"""
         self.driver.close()
 
     def __enter__(self) -> BrowserContext:
